from datetime import datetime
from typing import List, Any

import aiohttp
import pytest
import pytest_mock

from hoyolabrssfeeds import feeds
from hoyolabrssfeeds import models
from hoyolabrssfeeds.loaders import AbstractFeedFileLoader
from hoyolabrssfeeds.writers import AbstractFeedFileWriter
from hoyolabrssfeeds.writers import JSONFeedFileWriter


def test_game_feed_was_updated(
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
) -> None:
    feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)

    assert not feed.was_updated


def test_same_path_warning(
    feed_meta: models.FeedMeta,
    mocked_loader: AbstractFeedFileLoader,
    json_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    writer: AbstractFeedFileWriter = JSONFeedFileWriter(json_feed_file_writer_config)
    duplicate_writers = [writer, writer]

    with pytest.warns(UserWarning, match="identical paths"):
        feeds.GameFeed(feed_meta, duplicate_writers, mocked_loader)


def test_from_config(feed_config: models.FeedConfig) -> None:
    game_feed = feeds.GameFeed.from_config(feed_config)

    assert game_feed._feed_meta == feed_config.feed_meta

    for writer in game_feed._feed_writers:
        assert issubclass(type(writer), AbstractFeedFileWriter)

    writer_configs = [writer.config for writer in game_feed._feed_writers]
    assert writer_configs == feed_config.writer_configs

    assert issubclass(type(game_feed._feed_loader), AbstractFeedFileLoader)
    assert game_feed._feed_loader.config == feed_config.loader_config


def test_from_config_no_loader(feed_config_no_loader: models.FeedConfig) -> None:
    game_feed = feeds.GameFeed.from_config(feed_config_no_loader)

    assert game_feed._feed_loader is not None
    assert issubclass(type(game_feed._feed_loader), AbstractFeedFileLoader)


def test_init_no_loader(
    feed_meta: models.FeedMeta,
    json_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    writers: List[AbstractFeedFileWriter] = [
        JSONFeedFileWriter(json_feed_file_writer_config)
    ]
    game_feed = feeds.GameFeed(feed_meta, writers)

    assert game_feed._feed_loader.config.feed_type == models.FeedType.JSON


async def test_category_feed_new_item(
    mocker: pytest_mock.MockFixture,
    client_session: aiohttp.ClientSession,
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
    feed_item: models.FeedItem,
) -> None:
    feed_meta.category_size = 2

    new_item = feed_item.copy()
    new_item.id += 1
    new_item.published = datetime.now()
    new_item.updated = datetime.now()

    mocked_metas = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_latest_item_metas",
        spec=True,
        return_value=[
            models.FeedItemMeta(id=new_item.id, last_modified=new_item.updated),
            models.FeedItemMeta(id=feed_item.id, last_modified=feed_item.updated),
        ],
    )

    mocked_item = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_feed_item",
        spec=True,
        return_value=new_item,
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)
    updated_feed = await game_feed._update_category_feed(
        client_session, models.FeedItemCategory.INFO, [feed_item]
    )

    mocked_metas.assert_awaited()
    mocked_metas.assert_called_once()

    # only the new item should be fetched
    mocked_item.assert_awaited()
    mocked_item.assert_called_once()

    assert game_feed.was_updated

    # since the new item should be the latest, it needs to be the first in list
    # -> indirect check of sorting
    assert updated_feed == [new_item, feed_item]


async def test_category_feed_updated_item(
    mocker: pytest_mock.MockFixture,
    client_session: aiohttp.ClientSession,
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
    feed_item: models.FeedItem,
) -> None:
    feed_meta.category_size = 2

    updated_item = feed_item.copy()
    updated_item.updated = datetime.now().astimezone()

    other_item = feed_item.copy()
    other_item.id += 1

    mocked_metas = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_latest_item_metas",
        spec=True,
        return_value=[
            models.FeedItemMeta(id=other_item.id, last_modified=other_item.updated),
            models.FeedItemMeta(id=updated_item.id, last_modified=updated_item.updated),
        ],
    )

    mocked_item = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_feed_item",
        spec=True,
        return_value=updated_item,
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)
    updated_feed = await game_feed._update_category_feed(
        client_session, models.FeedItemCategory.INFO, [other_item, feed_item]
    )

    mocked_metas.assert_awaited()
    mocked_metas.assert_called_once()

    # only the updated item should be fetched
    mocked_item.assert_awaited()
    mocked_item.assert_called_once()

    assert game_feed.was_updated

    # feed is sorted by ids even if an item was updated
    assert updated_feed == [other_item, updated_item]


async def test_category_feed_unchanged(
    mocker: pytest_mock.MockFixture,
    client_session: aiohttp.ClientSession,
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
    feed_item: models.FeedItem,
) -> None:
    mocked_metas = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_latest_item_metas",
        spec=True,
        return_value=[
            models.FeedItemMeta(id=feed_item.id, last_modified=feed_item.updated)
        ],
    )

    mocked_item = mocker.patch(
        "hoyolabrssfeeds.feeds.HoyolabNews.get_feed_item", spec=True
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)
    updated_feed = await game_feed._update_category_feed(
        client_session, models.FeedItemCategory.INFO, [feed_item]
    )

    mocked_metas.assert_awaited()
    mocked_metas.assert_called_once()

    mocked_item.assert_not_called()

    assert not game_feed.was_updated

    assert updated_feed == [feed_item]


async def test_create_feed(
    mocker: pytest_mock.MockFixture,
    feed_meta: models.FeedMeta,
    mocked_writers: List[Any],
    mocked_loader: Any,
    category_feeds: List[List[models.FeedItem]],
    combined_feed: List[models.FeedItem],
) -> None:
    mocked_update_feed = mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed._update_category_feed",
        spec=True,
        side_effect=category_feeds,
    )

    mocked_was_updated = mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed._was_updated",
        new_callable=mocker.PropertyMock,
        create=True,
        return_value=True,
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)

    # no client session given to test local session in method
    await game_feed.create_feed()

    mocked_loader.get_feed_items.assert_awaited()
    mocked_loader.get_feed_items.assert_called_once()

    mocked_update_feed.assert_awaited()
    mocked_update_feed.assert_called()

    mocked_was_updated.assert_called()

    for writer in mocked_writers:
        writer.write_feed.assert_called_with(feed_meta, combined_feed)


async def test_create_feed_unchanged(
    mocker: pytest_mock.MockFixture,
    client_session: aiohttp.ClientSession,
    feed_meta: models.FeedMeta,
    mocked_writers: List[Any],
    mocked_loader: Any,
    category_feeds: List[List[models.FeedItem]],
    combined_feed: List[models.FeedItem],
) -> None:
    # set known feed items from file
    mocked_loader.get_feed_items.return_value = combined_feed

    mocked_update_feed = mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed._update_category_feed",
        spec=True,
        side_effect=category_feeds,
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)

    await game_feed.create_feed(client_session)

    mocked_loader.get_feed_items.assert_awaited()
    mocked_loader.get_feed_items.assert_called_once()

    mocked_update_feed.assert_awaited()
    mocked_update_feed.assert_called()

    for writer in mocked_writers:
        writer.write_feed.assert_not_called()


async def test_create_feed_one_category(
    mocker: pytest_mock.MockFixture,
    client_session: aiohttp.ClientSession,
    feed_meta: models.FeedMeta,
    feed_item: models.FeedItem,
    mocked_writers: List[Any],
    mocked_loader: Any,
) -> None:
    feed_meta.categories = [feed_item.category]

    mocked_update_feed = mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed._update_category_feed",
        spec=True,
        return_value=[feed_item],
    )

    # needed for writers to trigger
    mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed._was_updated",
        new_callable=mocker.PropertyMock,
        create=True,
        return_value=True,
    )

    game_feed = feeds.GameFeed(feed_meta, mocked_writers, mocked_loader)

    await game_feed.create_feed(client_session)

    mocked_update_feed.assert_awaited()
    mocked_update_feed.assert_called_once()

    for writer in mocked_writers:
        writer.write_feed.assert_called_with(feed_meta, [feed_item])


def test_collection_from_config(
    feed_config: models.FeedConfig, feed_config_no_loader: models.FeedConfig
) -> None:
    # NOTE: there is currently no check for identical paths of multiple game feeds
    configs = [feed_config, feed_config_no_loader]

    collection = feeds.GameFeedCollection.from_configs(configs)

    assert len(collection._game_feeds) == len(configs)

    for feed, config in zip(collection._game_feeds, configs):
        assert feed._feed_meta == config.feed_meta

        writer_configs = [w.config for w in feed._feed_writers]
        assert writer_configs == config.writer_configs

        if config.loader_config is not None:
            assert feed._feed_loader.config == config.loader_config


async def test_create_feed_collections(
    mocker: pytest_mock.MockFixture,
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
) -> None:
    mocked_create = mocker.patch(
        "hoyolabrssfeeds.feeds.GameFeed.create_feed", spec=True
    )

    collection = feeds.GameFeedCollection(
        [feed_meta, feed_meta],
        [mocked_writers, mocked_writers],
        [mocked_loader, mocked_loader],
    )

    await collection.create_feeds()

    mocked_create.assert_awaited()
    mocked_create.assert_called()


async def test_invalid_feed_collection(
    feed_meta: models.FeedMeta,
    mocked_writers: List[AbstractFeedFileWriter],
    mocked_loader: AbstractFeedFileLoader,
) -> None:
    with pytest.raises(ValueError):
        feeds.GameFeedCollection(
            [feed_meta, feed_meta],
            [mocked_writers],
            [mocked_loader, mocked_loader, mocked_loader],
        )



from pathlib import Path
from platform import system
from stat import S_IREAD
from stat import S_IWRITE
from typing import Any
from typing import Dict

import aiofiles
import pytest
import pytest_mock

try:
    import tomllib as toml
except ImportError:
    import tomli as toml  # type: ignore

from hoyolabrssfeeds import configs
from hoyolabrssfeeds import errors
from hoyolabrssfeeds import models


def test_config_paths(config_path: Path) -> None:
    param_config = configs.FeedConfigLoader(config_path)
    assert param_config.path == config_path

    fallback_config = configs.FeedConfigLoader()
    assert fallback_config.path == Path("hoyolab-rss-feeds.toml")


async def test_load_toml_file(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    toml_str = """
    [genshin]
    key = 'value'
    """

    async with aiofiles.open(config_path, "w") as fd:
        await fd.write(toml_str)

    config_dict = await loader._load_from_file()
    expected = {"genshin": {"key": "value"}}

    assert config_dict == expected


async def test_invalid_toml_file(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    async with aiofiles.open(config_path, "w") as fd:
        await fd.write("{Invalid}")

    with pytest.raises(errors.ConfigFormatError, match="Could not parse"):
        await loader._load_from_file()


async def test_empty_toml_file(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    config_path.touch()

    with pytest.raises(errors.ConfigFormatError, match="Could not find"):
        await loader._load_from_file()


async def test_default_toml_file(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    await loader.create_default_config_file()

    assert config_path.exists()

    async with aiofiles.open(config_path, "r") as fd:
        toml_str = await fd.read()

    default_config = toml.loads(toml_str)

    expected = {
        "genshin": {"feed": {"json": {"path": "genshin.json"}}},
    }

    assert default_config == expected


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_default_toml_file_io_error(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    # create read-only file
    config_path.touch(S_IREAD)

    with pytest.raises(errors.ConfigIOError, match="Could not create"):
        await loader.create_default_config_file()


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_load_toml_file_io_error(config_path: Path) -> None:
    loader = configs.FeedConfigLoader(config_path)

    # create write-only file
    config_path.touch(S_IWRITE)

    with pytest.raises(errors.ConfigIOError, match="Could not open"):
        await loader._load_from_file()


async def test_create_feed_config(
    mocker: pytest_mock.MockFixture,
    feed_config_no_loader: models.FeedConfig,
    toml_config_dict: Dict[str, Any],
) -> None:
    loader = configs.FeedConfigLoader()

    mocked_load = mocker.patch(
        "hoyolabrssfeeds.configs.FeedConfigLoader._load_from_file",
        spec=True,
        return_value=toml_config_dict,
    )

    loaded_config = await loader.get_feed_config(feed_config_no_loader.feed_meta.game)

    mocked_load.assert_awaited()
    mocked_load.assert_called_once()

    assert loaded_config == feed_config_no_loader


async def test_create_all_feed_configs(
    mocker: pytest_mock.MockFixture, toml_config_dict: Dict[str, Any]
) -> None:
    loader = configs.FeedConfigLoader()

    mocked_load = mocker.patch(
        "hoyolabrssfeeds.configs.FeedConfigLoader._load_from_file",
        spec=True,
        return_value=toml_config_dict,
    )

    loaded_configs = await loader.get_all_feed_configs()

    mocked_load.assert_awaited()
    mocked_load.assert_called_once()

    for conf in loaded_configs:
        # only testing if a feed config exists for all specified games
        assert conf.feed_meta.game.name.lower() in toml_config_dict


def test_create_invalid_feed_config() -> None:
    loader = configs.FeedConfigLoader()

    with pytest.raises(errors.ConfigFormatError, match="Could not find"):
        loader._create_feed_config(models.Game.GENSHIN, {"honkai": {}})

    with pytest.raises(errors.ConfigFormatError, match="Invalid config"):
        loader._create_feed_config(
            models.Game.GENSHIN, {"genshin": {"feed": {"Invalid": {}}}}
        )



import asyncio
import re
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import AsyncGenerator
from typing import Dict
from typing import List
from unittest.mock import MagicMock
from xml.etree import ElementTree

import aiohttp
import pydantic
import pytest
import pytest_mock

from hoyolabrssfeeds import models
from hoyolabrssfeeds.loaders import AbstractFeedFileLoader
from hoyolabrssfeeds.writers import AbstractFeedFileWriter


# ---- GENERAL FIXTURES ----


@pytest.fixture()
async def client_session() -> AsyncGenerator[aiohttp.ClientSession, Any]:
    loop = asyncio.get_running_loop()
    async with aiohttp.ClientSession(loop=loop, raise_for_status=True) as cs:
        yield cs


# ---- PATH FIXTURES ----


@pytest.fixture
def json_path(tmp_path: Path) -> Path:
    return tmp_path / Path("json_feed.json")


@pytest.fixture
def atom_path(tmp_path: Path) -> Path:
    return tmp_path / Path("atom_feed.xml")


@pytest.fixture
def config_path(tmp_path: Path) -> Path:
    return tmp_path / Path("feeds.toml")


# ---- MODEL FIXTURES ----


@pytest.fixture(
    params=[g for g in models.Game], ids=[g.name.lower() for g in models.Game]
)
def game(request: Any) -> Any:
    return request.param


@pytest.fixture(
    params=[c for c in models.FeedItemCategory],
    ids=[c.name.lower() for c in models.FeedItemCategory],
)
def category(request: Any) -> Any:
    return request.param


@pytest.fixture(params=[la for la in models.Language])
def language(request: Any) -> Any:
    return request.param


# ---- CONFIG FIXTURES ----


@pytest.fixture
def json_feed_file_writer_config(json_path: Path) -> models.FeedFileWriterConfig:
    return models.FeedFileWriterConfig(
        feed_type=models.FeedType.JSON,
        path=json_path,
        url=pydantic.parse_obj_as(pydantic.HttpUrl, "https://example.org/"),
    )


@pytest.fixture
def atom_feed_file_writer_config(atom_path: Path) -> models.FeedFileWriterConfig:
    return models.FeedFileWriterConfig(
        feed_type=models.FeedType.ATOM,
        path=atom_path,
        url=pydantic.parse_obj_as(pydantic.HttpUrl, "https://example.org/"),
    )


@pytest.fixture
def json_feed_file_config(json_path: Path) -> models.FeedFileConfig:
    return models.FeedFileConfig(feed_type=models.FeedType.JSON, path=json_path)


@pytest.fixture
def atom_feed_file_config(atom_path: Path) -> models.FeedFileConfig:
    return models.FeedFileConfig(feed_type=models.FeedType.ATOM, path=atom_path)


@pytest.fixture
def feed_config(
    feed_meta: models.FeedMeta,
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    json_feed_file_config: models.FeedFileConfig,
) -> models.FeedConfig:
    return models.FeedConfig(
        feed_meta=feed_meta,
        writer_configs=[json_feed_file_writer_config],
        loader_config=json_feed_file_config,
    )


@pytest.fixture
def feed_config_no_loader(
    feed_meta: models.FeedMeta,
    json_feed_file_writer_config: models.FeedFileWriterConfig,
) -> models.FeedConfig:
    return models.FeedConfig(
        feed_meta=feed_meta,
        writer_configs=[json_feed_file_writer_config],
    )


@pytest.fixture
def toml_config_dict(
    feed_meta: models.FeedMeta,
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
) -> Dict[str, Any]:
    return {
        "category_size": feed_meta.category_size,
        "language": str(feed_meta.language),
        feed_meta.game.name.lower(): {
            "feed": {
                str(models.FeedType.JSON): {
                    "path": str(json_feed_file_writer_config.path),
                    "url": str(json_feed_file_writer_config.url),
                }
            },
            "title": feed_meta.title,
            "icon": str(feed_meta.icon),
        },
        models.Game.ZENLESS.name.lower(): {
            "feed": {
                str(models.FeedType.ATOM): {
                    "path": str(atom_feed_file_writer_config.path)
                }
            },
            "categories": [models.FeedItemCategory.EVENTS.name.lower()],
        },
    }


# ---- FEED FIXTURES ----


@pytest.fixture
def feed_item() -> models.FeedItem:
    return models.FeedItem(
        id=42,
        title="Test Article",
        author="John Doe",
        content="<p>Hello World!</p>",
        category=models.FeedItemCategory.INFO,
        published=datetime(2022, 10, 3, 16).astimezone(),
        updated=datetime(2022, 10, 3, 18).astimezone(),
        image=pydantic.parse_obj_as(pydantic.HttpUrl, "https://example.org/"),
    )


@pytest.fixture
def feed_item_list(feed_item: models.FeedItem) -> List[models.FeedItem]:
    other_item = feed_item.copy()
    other_item.id -= 1
    return [feed_item, other_item]


@pytest.fixture
def json_feed_items(feed_item_list: List[models.FeedItem]) -> Dict[str, Any]:
    return {
        "items": [
            {
                "id": feed_item.id,
                "url": "https://example.org",
                "title": feed_item.title,
                "authors": [{"name": feed_item.author}],
                "tags": [feed_item.category.name.title()],
                "content_html": feed_item.content,
                "date_published": feed_item.published.astimezone().isoformat(),
                "date_modified": feed_item.updated.astimezone().isoformat()
                if feed_item.updated
                else "",
                "image": feed_item.image,
            }
            for feed_item in feed_item_list
        ]
    }


@pytest.fixture
def atom_feed_entries(feed_item_list: List[models.FeedItem]) -> ElementTree.Element:
    # omitting namespace declarations because they should be removed before
    root = ElementTree.Element("feed")

    for feed_item in feed_item_list:
        entry = ElementTree.SubElement(root, "entry")

        id_str = "tag:hoyolab.com,{}:{}".format(
            feed_item.published.date().isoformat(), feed_item.id
        )
        ElementTree.SubElement(entry, "id").text = id_str

        ElementTree.SubElement(entry, "title").text = feed_item.title
        ElementTree.SubElement(entry, "content").text = feed_item.content
        ElementTree.SubElement(
            entry, "category", {"term": feed_item.category.name.title()}
        )

        author = ElementTree.SubElement(entry, "author")
        ElementTree.SubElement(author, "name").text = feed_item.author

        published_str = feed_item.published.astimezone().isoformat()
        ElementTree.SubElement(entry, "published").text = published_str

        if feed_item.updated is not None:
            updated_str = feed_item.updated.astimezone().isoformat()
            ElementTree.SubElement(entry, "updated").text = updated_str

    return root


@pytest.fixture
def feed_meta() -> models.FeedMeta:
    return models.FeedMeta(
        game=models.Game.GENSHIN,
        category_size=1,
        categories=[c for c in models.FeedItemCategory],
        language=models.Language.GERMAN,
        title="Example Feed",
        icon=pydantic.parse_obj_as(pydantic.HttpUrl, "https://example.org/"),
    )


@pytest.fixture
def category_feeds(feed_item: models.FeedItem) -> List[List[models.FeedItem]]:
    cat_feeds: List[List[models.FeedItem]] = []
    for i, cat in enumerate(models.FeedItemCategory):
        item = feed_item.copy()
        item.id += i
        item.category = cat
        cat_feeds.append([item])

    return cat_feeds


@pytest.fixture
def combined_feed(category_feeds: List[List[models.FeedItem]]) -> List[models.FeedItem]:
    comb_feed: List[models.FeedItem] = []
    for item_list in category_feeds:
        comb_feed.extend(item_list)
    comb_feed.sort(key=lambda x: x.id, reverse=True)

    return comb_feed


# ---- MOCK FIXTURES ----


@pytest.fixture
def mocked_loader(mocker: pytest_mock.MockFixture) -> MagicMock:
    loader: MagicMock = mocker.create_autospec(AbstractFeedFileLoader, instance=True)
    loader.get_feed_items = mocker.AsyncMock(return_value=[])

    return loader


@pytest.fixture
def mocked_writers(mocker: pytest_mock.MockFixture) -> List[MagicMock]:
    writer: MagicMock = mocker.create_autospec(AbstractFeedFileWriter, instance=True)

    return [writer]


# ---- ASSERTION HELPERS ----

# https://docs.pytest.org/en/6.2.x/assert.html#assertion-introspection-details


def validate_hoyolab_post(post: Dict[str, Any], is_full_post: bool) -> None:
    assert type(post["post"]["post_id"]) is str
    assert re.fullmatch(r"\d+", post["post"]["post_id"]) is not None

    assert type(post["post"]["created_at"]) is int
    assert post["post"]["created_at"] > 0

    assert type(post["last_modify_time"]) is int
    assert post["last_modify_time"] >= 0

    if is_full_post:
        assert type(post["post"]["official_type"]) is int
        assert post["post"]["official_type"] in [
            c.value for c in models.FeedItemCategory
        ]

        assert type(post["user"]["nickname"]) is str
        assert len(post["user"]["nickname"]) > 0

        assert type(post["post"]["content"]) is str
        assert len(post["post"]["content"]) > 0

        assert type(post["post"]["subject"]) is str
        assert len(post["post"]["subject"]) > 0

        assert type(post["cover_list"]) is list
        assert len(post["cover_list"]) >= 0



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import aiofiles
import pydantic

try:
    import tomllib as toml
except ImportError:
    import tomli as toml  # type: ignore

from .errors import ConfigIOError
from .errors import ConfigFormatError
from .models import FeedConfig
from .models import FeedFileWriterConfig
from .models import FeedItemCategory
from .models import FeedMeta
from .models import Game


class FeedConfigLoader:
    """TOML config file loader."""

    def __init__(self, config_path: Optional[Path] = None):
        self._path = config_path or Path("hoyolab-rss-feeds.toml")

    @property
    def path(self) -> Path:
        """TOML config file path."""
        return self._path

    async def _load_from_file(self) -> Dict[str, Any]:
        """Load and parse config from TOML file."""

        try:
            async with aiofiles.open(self._path, "r") as fd:
                conf_str = await fd.read()

            config: Dict[str, Any] = toml.loads(conf_str)
        except IOError as err:
            raise ConfigIOError(
                'Could not open config file at "{}"!'.format(self._path)
            ) from err
        except toml.TOMLDecodeError as err:
            raise ConfigFormatError("Could not parse TOML config file!") from err

        if not any([game.name.lower() in config for game in Game]):
            raise ConfigFormatError("Could not find any game configs!")

        return config

    @staticmethod
    def _create_feed_config(game: Game, config_dict: Dict[str, Any]) -> FeedConfig:
        """Create a feed config from a TOML dict for a specified game."""

        games = {g.name.lower() for g in Game}

        try:
            game_config_dict = config_dict[game.name.lower()]

            # merge root keys into game config dict
            for key, val in config_dict.items():
                if key not in games and key != "feed":
                    # only set key if not already exists
                    game_config_dict.setdefault(key, val)

            feed_config_dict = game_config_dict.pop("feed")

            writer_configs = [
                FeedFileWriterConfig(feed_type=feed_type, **feed_config)
                for feed_type, feed_config in feed_config_dict.items()
            ]

            if "categories" in game_config_dict:
                game_config_dict["categories"] = list(
                    map(
                        lambda cat: FeedItemCategory.from_str(cat),
                        game_config_dict["categories"],
                    )
                )

            feed_meta = FeedMeta(game=game, **game_config_dict)
            feed_config = FeedConfig(feed_meta=feed_meta, writer_configs=writer_configs)
        except KeyError as err:
            raise ConfigFormatError("Could not find required key in config!") from err
        except (pydantic.ValidationError, ValueError) as err:
            raise ConfigFormatError("Invalid config value!") from err

        return feed_config

    async def get_feed_config(self, game: Game) -> FeedConfig:
        """Load and create a feed config for a given game."""

        config = await self._load_from_file()

        return self._create_feed_config(game, config)

    async def get_all_feed_configs(self) -> List[FeedConfig]:
        """Load and create feed configs for all games found in file."""

        config = await self._load_from_file()

        return [
            self._create_feed_config(Game.from_str(key), config)
            for key in config.keys()
            if key in {g.name.lower() for g in Game}
        ]

    async def create_default_config_file(self) -> None:
        """Create an initial example config file."""

        toml_str = '[genshin]\nfeed.json.path = "genshin.json"'

        try:
            async with aiofiles.open(self._path, "w") as fd:
                await fd.write(toml_str)
        except IOError as err:
            raise ConfigIOError(
                'Could not create default config file at "{}"!'.format(self._path)
            ) from err



"""RSS feed generator for official game news from Hoyolab."""

from . import configs
from . import errors
from . import feeds
from . import hoyolab
from . import loaders
from . import models
from . import writers

# quick access
from .configs import FeedConfigLoader
from .feeds import GameFeed
from .feeds import GameFeedCollection
from .models import Game

__all__ = [
    "configs",
    "errors",
    "feeds",
    "hoyolab",
    "loaders",
    "models",
    "writers",
    "FeedConfigLoader",
    "GameFeed",
    "GameFeedCollection",
    "Game",
]



from datetime import datetime
from datetime import timezone

import aiohttp
import langdetect  # type: ignore
import pydantic
import pytest
import pytest_mock

from hoyolabrssfeeds import errors
from hoyolabrssfeeds import hoyolab
from hoyolabrssfeeds import models
from .conftest import validate_hoyolab_post

# force consistent results
langdetect.DetectorFactory.seed = 0


@pytest.mark.hoyolabapi
async def test_api_post_endpoint(
    client_session: aiohttp.ClientSession, game: models.Game
) -> None:
    post_id = get_post_id(game)
    api = hoyolab.HoyolabNews(game)
    post = await api.get_post(client_session, post_id)

    validate_hoyolab_post(post, is_full_post=True)


@pytest.mark.hoyolabapi
async def test_api_list_endpoint(
    client_session: aiohttp.ClientSession,
    game: models.Game,
    category: models.FeedItemCategory,
) -> None:
    api = hoyolab.HoyolabNews(game)
    news_list = await api.get_news_list(client_session, category, 2)

    assert type(news_list) is list
    assert len(news_list) > 0

    for post in news_list:
        validate_hoyolab_post(post, is_full_post=False)


@pytest.mark.hoyolabapi
@pytest.mark.xfail(reason="not accurate for all languages")
async def test_language(
    client_session: aiohttp.ClientSession, language: models.Language
) -> None:
    # NOTE: The list endpoint is not checked due to insufficient text.
    # The text of the list endpoint is also not used for the feeds.

    api = hoyolab.HoyolabNews(models.Game.GENSHIN, language)

    post_id = 14189622
    post = await api.get_post(client_session, post_id)
    content = post["post"]["content"]

    # only use first letters of code if not chinese
    # https://github.com/Mimino666/langdetect
    expected = language.lower()
    if not expected.startswith("zh"):
        expected = expected.partition("-")[0]

    assert langdetect.detect(content) == expected


@pytest.mark.hoyolabapi
async def test_request_errors(client_session: aiohttp.ClientSession) -> None:
    api = hoyolab.HoyolabNews(models.Game.GENSHIN)

    # request error
    error_url = pydantic.parse_obj_as(
        pydantic.HttpUrl, "https://httpbin.org/status/500"
    )
    with pytest.raises(errors.HoyolabApiError, match="Could not request"):
        await api._request(client_session, {}, error_url)

    # decode error
    error_url = pydantic.parse_obj_as(pydantic.HttpUrl, "https://httpbin.org/html")
    with pytest.raises(errors.HoyolabApiError, match="Could not decode"):
        await api._request(client_session, {}, error_url)

    # unexpected response
    error_url = pydantic.parse_obj_as(pydantic.HttpUrl, "https://httpbin.org/json")
    with pytest.raises(errors.HoyolabApiError, match="Unexpected response"):
        await api._request(client_session, {}, error_url)

    # hoyolab error code
    error_url = pydantic.parse_obj_as(
        pydantic.HttpUrl,
        "https://bbs-api-os.hoyolab.com/community/post/wapi/getNewsList",
    )
    with pytest.raises(errors.HoyolabApiError):
        # missing params raise exception here
        await api._request(client_session, {}, error_url)


async def test_get_latest_item_metas(
    mocker: pytest_mock.MockFixture, client_session: aiohttp.ClientSession
) -> None:
    posts = [
        {"post": {"post_id": "42", "created_at": 1645564944}, "last_modify_time": 0},
        {
            "post": {"post_id": "41", "created_at": 1645564942},
            "last_modify_time": 1645564943,
        },
    ]

    mocked_news_list = mocker.patch(
        "hoyolabrssfeeds.hoyolab.HoyolabNews.get_news_list",
        spec=True,
        return_value=posts,
    )

    api = hoyolab.HoyolabNews(models.Game.GENSHIN)
    metas = await api.get_latest_item_metas(
        client_session, models.FeedItemCategory.INFO, 2
    )

    expected = [
        models.FeedItemMeta(
            id=42, last_modified=datetime.fromtimestamp(1645564944, tz=timezone.utc)
        ),
        models.FeedItemMeta(
            id=41, last_modified=datetime.fromtimestamp(1645564943, tz=timezone.utc)
        ),
    ]

    mocked_news_list.assert_awaited()
    mocked_news_list.assert_called()

    assert metas == expected


async def test_get_feed_item(
    mocker: pytest_mock.MockFixture,
    feed_item: models.FeedItem,
    client_session: aiohttp.ClientSession,
) -> None:
    post = {
        "post": {
            "post_id": str(feed_item.id),
            "subject": feed_item.title,
            "content": feed_item.content,
            "official_type": feed_item.category.value,
            "created_at": int(feed_item.published.timestamp()),
        },
        "user": {"nickname": feed_item.author},
        "last_modify_time": int(
            feed_item.updated.timestamp() if feed_item.updated else 0
        ),
        "cover_list": [{"url": str(feed_item.image)}],
    }

    mocked_get_post = mocker.patch(
        "hoyolabrssfeeds.hoyolab.HoyolabNews.get_post", spec=True, return_value=post
    )

    api = hoyolab.HoyolabNews(models.Game.GENSHIN)
    fetched_item = await api.get_feed_item(client_session, feed_item.id)

    mocked_get_post.assert_awaited()
    mocked_get_post.assert_called()

    assert fetched_item == feed_item


def test_leading_line_breaks() -> None:
    post = {"post": {"content": "<p><br></p>Hello World"}}

    expected = {"post": {"content": "Hello World"}}

    transformed_post = hoyolab.HoyolabNews._transform_post(post)

    assert transformed_post == expected


def test_private_link_bug() -> None:
    post = {
        "post": {
            "content": '<img src="https://hoyolab-upload-private.hoyolab.com/test.jpg">'
        }
    }

    expected = {
        "post": {"content": '<img src="https://upload-os-bbs.hoyolab.com/test.jpg">'}
    }

    transformed_post = hoyolab.HoyolabNews._transform_post(post)

    assert transformed_post == expected


def test_content_html_bug() -> None:
    post = {"post": {"content": "en-us"}}

    transformed_post = hoyolab.HoyolabNews._transform_post(post)

    # test that the content was replaced/fixed
    assert transformed_post["post"]["content"] != "en-us"


# ---- HELPER FUNCTIONS ----


def get_post_id(game: models.Game) -> int:
    post_ids = {
        models.Game.HONKAI: 4361615,
        models.Game.GENSHIN: 7156359,
        models.Game.THEMIS: 4283335,
        models.Game.STARRAIL: 3746616,
        models.Game.ZENLESS: 4729212,
    }

    return post_ids[game]



import asyncio
from pathlib import Path

import aiofiles
import aiohttp

from hoyolabrssfeeds import FeedConfigLoader
from hoyolabrssfeeds import Game
from hoyolabrssfeeds import GameFeed
from hoyolabrssfeeds import GameFeedCollection


async def test_single_feed(
    client_session: aiohttp.ClientSession, tmp_path: Path
) -> None:
    config_path = await _write_config(tmp_path)
    config_loader = FeedConfigLoader(config_path)
    game_config = await config_loader.get_feed_config(Game.GENSHIN)
    game_feed = GameFeed.from_config(game_config)

    # initial creation
    await game_feed.create_feed(client_session)

    for feed_path in [w.path for w in game_config.writer_configs]:
        assert feed_path.exists()

    await asyncio.sleep(2)

    # feed update
    await game_feed.create_feed(client_session)

    assert not game_feed.was_updated


async def test_feed_collection(
    client_session: aiohttp.ClientSession, tmp_path: Path
) -> None:
    config_path = await _write_config(tmp_path)
    config_loader = FeedConfigLoader(config_path)
    game_configs = await config_loader.get_all_feed_configs()

    feed_collection = GameFeedCollection.from_configs(game_configs)
    await feed_collection.create_feeds(client_session)

    for conf in game_configs:
        for w in conf.writer_configs:
            assert w.path.exists()


# ---- HELPER FUNCTIONS ----


async def _write_config(base_tmp_path: Path) -> Path:
    # NOTE: paths are single quoted to get literal strings
    # this should avoid wrongfully escaping of windows paths by tomli
    toml_templ = """
        icon = "https://example.org"

        [genshin]
        feed.json.path = '{}'
        feed.json.url = "https://example.org"
        feed.atom.path = '{}'
        title = "Genshin"

        [honkai]
        feed.atom.path = '{}'
        category_size = 3
    """

    toml_config = toml_templ.format(
        base_tmp_path / Path("genshin.json"),
        base_tmp_path / Path("genshin.xml"),
        base_tmp_path / Path("honkai.xml"),
    )

    config_path = base_tmp_path / Path("feeds.toml")
    async with aiofiles.open(config_path, "w") as fd:
        await fd.write(toml_config)

    return config_path



import re
from typing import Any
from typing import Dict
from typing import List

import aiohttp
import pydantic

from .errors import HoyolabApiError
from .models import FeedItem
from .models import FeedItemCategory
from .models import FeedItemMeta
from .models import Game
from .models import Language

HOYOLAB_API_BASE_URL = "https://bbs-api-os.hoyolab.com/community/post/wapi/"
DEFAULT_CATEGORY_SIZE = 5


class HoyolabNews:
    """Wrapper for Hoyolab REST API endpoints."""

    def __init__(self, game: Game, language: Language = Language.ENGLISH) -> None:
        self._game = game
        self._lang = language.lower()

    async def _request(
        self,
        session: aiohttp.ClientSession,
        params: Dict[str, Any],
        url: pydantic.HttpUrl,
    ) -> Dict[str, Any]:
        """Send a GET request to the Hoyolab API endpoint."""

        headers = {"Origin": "https://www.hoyolab.com", "X-Rpc-Language": self._lang}

        try:
            async with session.get(
                str(url), headers=headers, params=params
            ) as response:
                response.raise_for_status()
                response_json: Dict[str, Any] = await response.json()

            if response_json["retcode"] != 0:
                # the message might be in chinese
                raise HoyolabApiError(response_json["message"])
        except aiohttp.ContentTypeError as err:
            raise HoyolabApiError("Could not decode response to JSON!") from err
        except aiohttp.ClientResponseError as err:
            raise HoyolabApiError("Could not request Hoyolab endpoint!") from err
        except KeyError as err:
            raise HoyolabApiError("Unexpected response!") from err

        return response_json

    @staticmethod
    def _transform_post(post: Dict[str, Any]) -> Dict[str, Any]:
        """Transform (i.e. apply fixes) post of Hoyolab API response."""

        # remove empty leading paragraphs
        if post["post"]["content"].startswith(
            ("<p></p>", "<p>&nbsp;</p>", "<p><br></p>")
        ):
            post["post"]["content"] = post["post"]["content"].partition("</p>")[2]

        # fix private links
        post["post"]["content"] = post["post"]["content"].replace(
            "hoyolab-upload-private", "upload-os-bbs"
        )

        # weird hoyolab bug/feature, where the content html is just a language code.
        # could be fixed by parsing the structured_content and creating html from it.
        if re.fullmatch(r"^[a-z]{2}-[a-z]{2}$", post["post"]["content"]):
            post["post"]["content"] = "<em>Content not available...</em>"

        return post

    async def get_news_list(
        self,
        session: aiohttp.ClientSession,
        category: FeedItemCategory,
        category_size: int = DEFAULT_CATEGORY_SIZE,
    ) -> List[Dict[str, Any]]:
        """Request an overview of the latest posts."""

        params = {"gids": self._game, "page_size": category_size, "type": category}

        url = pydantic.parse_obj_as(
            pydantic.HttpUrl, HOYOLAB_API_BASE_URL + "getNewsList"
        )

        response = await self._request(session, params, url)
        news_list: List[Dict[str, Any]] = response["data"]["list"]

        return news_list

    async def get_post(
        self, session: aiohttp.ClientSession, post_id: int
    ) -> Dict[str, Any]:
        """Request single post with fulltext."""

        params = {"gids": self._game, "post_id": post_id}

        url = pydantic.parse_obj_as(
            pydantic.HttpUrl, HOYOLAB_API_BASE_URL + "getPostFull"
        )

        response = await self._request(session, params, url)
        post: Dict[str, Any] = response["data"]["post"]

        return self._transform_post(post)

    async def get_latest_item_metas(
        self,
        session: aiohttp.ClientSession,
        category: FeedItemCategory,
        category_size: int = DEFAULT_CATEGORY_SIZE,
    ) -> List[FeedItemMeta]:
        """Get the meta info of the latest posts in a specified category."""

        latest_posts = []

        category_posts = await self.get_news_list(session, category, category_size)

        for post in category_posts:
            published_ts = int(post["post"]["created_at"])
            modified_ts = int(post["last_modify_time"])

            item_meta = {
                "id": post["post"]["post_id"],
                "last_modified": max(published_ts, modified_ts),
            }

            # parsing for type conversions
            latest_posts.append(pydantic.parse_obj_as(FeedItemMeta, item_meta))

        return latest_posts

    async def get_feed_item(
        self, session: aiohttp.ClientSession, post_id: int
    ) -> FeedItem:
        """Get a single post as feed item."""

        post = await self.get_post(session, post_id)

        item = {
            "id": post["post"]["post_id"],
            "title": post["post"]["subject"],
            "author": post["user"]["nickname"],
            "content": post["post"]["content"],
            "category": post["post"]["official_type"],
            "published": post["post"]["created_at"],
        }

        if post["last_modify_time"] > 0:
            item["updated"] = post["last_modify_time"]

        if len(post["cover_list"]) > 0:
            item["image"] = post["cover_list"][0]["url"]

        return pydantic.parse_obj_as(FeedItem, item)



class HoyolabRssFeedsBaseError(Exception):
    """Base Error for this package."""


class ConfigIOError(HoyolabRssFeedsBaseError):
    """Raised if an IO operation on a config file failed."""


class FeedIOError(HoyolabRssFeedsBaseError):
    """Raised if an IO operation on a feed file failed."""


class HoyolabApiError(HoyolabRssFeedsBaseError):
    """Raised if interaction with the Hoyolab API failed."""


class ConfigFormatError(HoyolabRssFeedsBaseError):
    """Raised if an invalid config syntax or value is found."""


class FeedFormatError(HoyolabRssFeedsBaseError):
    """Raised if an invalid feed syntax or value is found."""



import json
from platform import system
from stat import S_IWRITE
from typing import Any
from typing import Dict
from typing import List
from xml.etree import ElementTree

import aiofiles
import pytest
import pytest_mock

from hoyolabrssfeeds import errors
from hoyolabrssfeeds import loaders
from hoyolabrssfeeds import models
from hoyolabrssfeeds import writers


# ---- FACTORY TESTS ----


def test_factory_feed_types() -> None:
    factory = loaders.FeedFileLoaderFactory()
    expected = {models.FeedType.ATOM, models.FeedType.JSON}

    assert factory.feed_types == expected


def test_factory_create_loader(
    json_feed_file_config: models.FeedFileConfig,
    atom_feed_file_config: models.FeedFileConfig,
) -> None:
    factory = loaders.FeedFileLoaderFactory()
    json_loader = factory.create_loader(json_feed_file_config)

    assert issubclass(type(json_loader), loaders.AbstractFeedFileLoader)
    assert isinstance(json_loader, loaders.JSONFeedFileLoader)

    atom_loader = factory.create_loader(atom_feed_file_config)

    assert issubclass(type(atom_loader), loaders.AbstractFeedFileLoader)
    assert isinstance(atom_loader, loaders.AtomFeedFileLoader)


def test_factory_create_any_loader(
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    json_writer = writers.JSONFeedFileWriter(json_feed_file_writer_config)
    atom_writer = writers.AtomFeedFileWriter(atom_feed_file_writer_config)

    factory = loaders.FeedFileLoaderFactory()
    json_loader = factory.create_any_loader([atom_writer, json_writer])

    assert issubclass(type(json_loader), loaders.AbstractFeedFileLoader)
    assert isinstance(json_loader, loaders.JSONFeedFileLoader)

    atom_loader = factory.create_any_loader([atom_writer])

    assert issubclass(type(atom_loader), loaders.AbstractFeedFileLoader)
    assert isinstance(atom_loader, loaders.AtomFeedFileLoader)


def test_factory_create_invalid_any_loader() -> None:
    factory = loaders.FeedFileLoaderFactory()

    with pytest.raises(ValueError, match="Could not create"):
        factory.create_any_loader([])


# ---- JSON LOADER TESTS ----


def test_json_feed_loader_config(json_feed_file_config: models.FeedFileConfig) -> None:
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    assert loader.config == json_feed_file_config


async def test_json_feed_loader(
    mocker: pytest_mock.MockFixture,
    json_feed_items: Dict[str, Any],
    feed_item_list: List[models.FeedItem],
    json_feed_file_config: models.FeedFileConfig,
) -> None:
    mocked_load = mocker.patch(
        "hoyolabrssfeeds.loaders.JSONFeedFileLoader._load_from_file",
        spec=True,
        return_value=json_feed_items,
    )

    # needed for get_feed_items() to work for this test
    json_feed_file_config.path.touch()

    loader = loaders.JSONFeedFileLoader(json_feed_file_config)
    loaded_items = await loader.get_feed_items()

    mocked_load.assert_awaited()
    mocked_load.assert_called()

    assert loaded_items == feed_item_list


async def test_invalid_json_feed_values(
    mocker: pytest_mock.MockFixture,
    json_feed_items: Dict[str, Any],
    json_feed_file_config: models.FeedFileConfig,
) -> None:
    # set invalid value
    json_feed_items["items"][0]["tags"] = ["invalid"]

    mocker.patch(
        "hoyolabrssfeeds.loaders.JSONFeedFileLoader._load_from_file",
        spec=True,
        return_value=json_feed_items,
    )

    json_feed_file_config.path.touch()
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    with pytest.raises(errors.FeedFormatError, match="Could not load"):
        await loader.get_feed_items()


async def test_invalid_json_feed(
    mocker: pytest_mock.MockFixture, json_feed_file_config: models.FeedFileConfig
) -> None:
    mocker.patch(
        "hoyolabrssfeeds.loaders.JSONFeedFileLoader._load_from_file",
        spec=True,
        return_value={},
    )

    json_feed_file_config.path.touch()
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    with pytest.raises(errors.FeedFormatError, match="Could not find"):
        await loader.get_feed_items()


async def test_no_json_file(json_feed_file_config: models.FeedFileConfig) -> None:
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    assert await loader.get_feed_items() == []


async def test_load_json_file(json_feed_file_config: models.FeedFileConfig) -> None:
    data = {"version": "https://jsonfeed.org/version/1.1"}

    async with aiofiles.open(json_feed_file_config.path, "w") as fd:
        await fd.write(json.dumps(data))

    loader = loaders.JSONFeedFileLoader(json_feed_file_config)
    loaded_data = await loader._load_from_file()

    assert loaded_data == data


async def test_load_invalid_json_file(
    json_feed_file_config: models.FeedFileConfig,
) -> None:
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    # write invalid json file
    async with aiofiles.open(json_feed_file_config.path, "w") as fd:
        await fd.write("Not JS0N!")

    with pytest.raises(errors.FeedFormatError, match="Could not decode"):
        await loader._load_from_file()


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_load_json_file_io_error(
    json_feed_file_config: models.FeedFileConfig,
) -> None:
    loader = loaders.JSONFeedFileLoader(json_feed_file_config)

    # create write only file
    json_feed_file_config.path.touch(S_IWRITE)

    with pytest.raises(errors.FeedIOError, match="Could not read"):
        await loader._load_from_file()


# ---- ATOM LOADER TESTS ----


def test_atom_feed_loader_config(atom_feed_file_config: models.FeedFileConfig) -> None:
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    assert loader.config == atom_feed_file_config


async def test_atom_feed_loader(
    mocker: pytest_mock.MockFixture,
    feed_item_list: List[models.FeedItem],
    atom_feed_entries: ElementTree.Element,
    atom_feed_file_config: models.FeedFileConfig,
) -> None:
    mocked_load = mocker.patch(
        "hoyolabrssfeeds.loaders.AtomFeedFileLoader._load_from_file",
        spec=True,
        return_value=atom_feed_entries,
    )

    # images cannot be stored in atom feeds
    for feed_item in feed_item_list:
        feed_item.image = None

    # needed for get_feed_items() to work for this test
    atom_feed_file_config.path.touch()
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    loaded_items = await loader.get_feed_items()

    mocked_load.assert_awaited()
    mocked_load.assert_called()

    assert loaded_items == feed_item_list


async def test_invalid_atom_feed_values(
    mocker: pytest_mock.MockFixture,
    atom_feed_entries: ElementTree.Element,
    atom_feed_file_config: models.FeedFileConfig,
) -> None:
    # set invalid value in first entry
    category_element = atom_feed_entries.find("entry/category")
    if category_element is not None:
        category_element.set("term", "invalid")
    else:
        pytest.fail(reason="Category element could not be found!")

    mocker.patch(
        "hoyolabrssfeeds.loaders.AtomFeedFileLoader._load_from_file",
        spec=True,
        return_value=atom_feed_entries,
    )

    atom_feed_file_config.path.touch()
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    with pytest.raises(errors.FeedFormatError, match="Could not load"):
        await loader.get_feed_items()


async def test_invalid_atom_feed(
    mocker: pytest_mock.MockFixture, atom_feed_file_config: models.FeedFileConfig
) -> None:
    invalid_feed = ElementTree.Element("feed")
    ElementTree.SubElement(invalid_feed, "entry")

    mocker.patch(
        "hoyolabrssfeeds.loaders.AtomFeedFileLoader._load_from_file",
        spec=True,
        return_value=invalid_feed,
    )

    atom_feed_file_config.path.touch()
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    with pytest.raises(errors.FeedFormatError, match="Could not load"):
        await loader.get_feed_items()


async def test_no_atom_file(atom_feed_file_config: models.FeedFileConfig) -> None:
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    assert await loader.get_feed_items() == []


async def test_load_atom_file(atom_feed_file_config: models.FeedFileConfig) -> None:
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    # faking namespace declaration because it is supposed to be removed in fn
    root = ElementTree.Element("feed", {"xmlns": "http://www.w3.org/2005/Atom"})

    title_text = "Test Atom Feed"
    ElementTree.SubElement(root, "title").text = title_text

    xml_bytes = ElementTree.tostring(root, encoding="utf-8")

    async with aiofiles.open(atom_feed_file_config.path, "wb") as fd:
        await fd.write(xml_bytes)

    loaded_feed = await loader._load_from_file()

    assert loaded_feed.findtext("title") == title_text


async def test_load_invalid_atom_file(
    atom_feed_file_config: models.FeedFileConfig,
) -> None:
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    async with aiofiles.open(atom_feed_file_config.path, "w") as fd:
        await fd.write("Not At0m!")

    with pytest.raises(errors.FeedFormatError, match="Could not parse"):
        await loader._load_from_file()


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_load_atom_file_io_error(
    atom_feed_file_config: models.FeedFileConfig,
) -> None:
    loader = loaders.AtomFeedFileLoader(atom_feed_file_config)

    # create write only file
    atom_feed_file_config.path.touch(S_IWRITE)

    with pytest.raises(errors.FeedIOError, match="Could not read"):
        await loader._load_from_file()



import json
from abc import ABCMeta
from abc import abstractmethod
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from xml.etree import ElementTree

import aiofiles
import pydantic

from .errors import FeedFormatError
from .errors import FeedIOError
from .models import FeedFileConfig
from .models import FeedItem
from .models import FeedItemCategory
from .models import FeedType
from .writers import AbstractFeedFileWriter
from .writers import JSONFeedFileWriter


class AbstractFeedFileLoader(metaclass=ABCMeta):
    """ABC for feed file loading functionality."""

    def __init__(self, config: FeedFileConfig) -> None:
        self._config = config

    @property
    def config(self) -> FeedFileConfig:
        """Returns the config of the feed loader."""
        return self._config

    @abstractmethod
    async def get_feed_items(self) -> List[FeedItem]:
        """Get the items of the feed or an empty list if they do not exist."""
        pass


class FeedFileLoaderFactory:
    """Factory for creating specific feed loaders."""

    def __init__(self) -> None:
        self._loaders = {
            FeedType.JSON: JSONFeedFileLoader,
            FeedType.ATOM: AtomFeedFileLoader,
        }

    @property
    def feed_types(self) -> Set[FeedType]:
        """Set of feed types for which writers are registered."""
        return set(self._loaders.keys())

    def create_loader(self, config: FeedFileConfig) -> AbstractFeedFileLoader:
        """Create feed loader for the specified feed type."""
        return self._loaders[config.feed_type](config)

    def create_any_loader(
        self, writers: List[AbstractFeedFileWriter]
    ) -> AbstractFeedFileLoader:
        """Create a suitable loader from given writers."""

        # prefer json loader if available
        for writer in writers:
            if isinstance(writer, JSONFeedFileWriter):
                json_config = pydantic.parse_obj_as(FeedFileConfig, writer.config)
                return self.create_loader(json_config)

        for writer in writers:
            if writer.config.feed_type in self._loaders:
                loader_config = pydantic.parse_obj_as(FeedFileConfig, writer.config)
                return self.create_loader(loader_config)

        raise ValueError("Could not create loader from given writers!")


class JSONFeedFileLoader(AbstractFeedFileLoader):
    """Load feed from JSON-Feed format (https://www.jsonfeed.org/version/1.1/)."""

    async def get_feed_items(self) -> List[FeedItem]:
        """Returns feed items of JSON-Feed if feed exists."""

        if self.config.path.exists():
            feed_items = []
            feed = await self._load_from_file()

            try:
                for item in feed["items"]:
                    category = FeedItemCategory.from_str(item["tags"][0])

                    item_dict = {
                        "id": item["id"],
                        "title": item["title"],
                        "author": item["authors"][0]["name"],
                        "content": item["content_html"],
                        "category": category,
                        "published": item["date_published"],
                    }

                    if "date_modified" in item:
                        item_dict["updated"] = item["date_modified"]

                    if "image" in item:
                        item_dict["image"] = item["image"]

                    feed_items.append(item_dict)
            except KeyError as err:
                raise FeedFormatError(
                    "Could not find required key in JSON feed!"
                ) from err
            except ValueError as err:
                raise FeedFormatError("Could not load JSON feed items!") from err

            return pydantic.parse_obj_as(List[FeedItem], feed_items)
        else:
            return []

    async def _load_from_file(self) -> Dict[str, Any]:
        """Load JSON-Feed from file."""

        try:
            async with aiofiles.open(self.config.path, "r") as fd:
                feed_json = await fd.read()

            feed: Dict[str, Any] = json.loads(feed_json)
        except IOError as err:
            raise FeedIOError(
                'Could not read JSON file from "{}"!'.format(self.config.path)
            ) from err
        except json.JSONDecodeError as err:
            raise FeedFormatError("Could not decode JSON file!") from err

        return feed


class AtomFeedFileLoader(AbstractFeedFileLoader):
    """Load feed from Atom format (https://validator.w3.org/feed/docs/atom.html)."""

    async def get_feed_items(self) -> List[FeedItem]:
        """Returns feed items of Atom feed if feed exists."""

        if self.config.path.exists():
            feed_items = []
            root = await self._load_from_file()

            for entry in root.findall("entry"):
                id_str = entry.findtext("id")
                item_id = id_str.rpartition(":")[2] if id_str is not None else None

                category_node = entry.find("category")
                try:
                    category = (
                        FeedItemCategory.from_str(category_node.get("term", default=""))
                        if category_node is not None
                        else None
                    )
                except ValueError as err:
                    raise FeedFormatError("Could not load Atom feed entries!") from err

                published_str = entry.findtext("published")
                published = (
                    datetime.fromisoformat(published_str)
                    if published_str is not None
                    else None
                )

                updated_str = entry.findtext("updated")
                updated = (
                    datetime.fromisoformat(updated_str)
                    if updated_str is not None
                    else None
                )

                item_dict = {
                    "id": item_id,
                    "title": entry.findtext("title"),
                    "author": entry.findtext("author/name"),
                    "content": entry.findtext("content"),
                    "category": category,
                    "published": published,
                    "updated": updated,
                }

                feed_items.append(item_dict)

            try:
                return pydantic.parse_obj_as(List[FeedItem], feed_items)
            except pydantic.ValidationError as err:
                raise FeedFormatError("Could not load Atom feed entries!") from err
        else:
            return []

    async def _load_from_file(self) -> ElementTree.Element:
        """Load Atom feed from file."""

        try:
            async with aiofiles.open(self.config.path, "r") as fd:
                feed_str = await fd.read()

            # removing default namespace declaration from xml because it makes
            # parsing MUCH easier
            feed_str = feed_str.replace(' xmlns="http://www.w3.org/2005/Atom"', "", 1)

            root = ElementTree.fromstring(feed_str)
        except IOError as err:
            raise FeedIOError(
                'Could not read Atom file from "{}"!'.format(self.config.path)
            ) from err
        except ElementTree.ParseError as err:
            raise FeedFormatError("Could not parse Atom file!") from err

        return root



import json
from abc import ABCMeta
from abc import abstractmethod
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from xml.etree import ElementTree

import aiofiles

from .errors import FeedIOError
from .models import FeedFileWriterConfig
from .models import FeedItem
from .models import FeedMeta
from .models import FeedType


class AbstractFeedFileWriter(metaclass=ABCMeta):
    """ABC for feed file writing functionality."""

    def __init__(self, config: FeedFileWriterConfig) -> None:
        self._config = config

    @property
    def config(self) -> FeedFileWriterConfig:
        """Returns the config of feed writer."""
        return self._config

    @abstractmethod
    async def write_feed(self, feed_meta: FeedMeta, feed_items: List[FeedItem]) -> None:
        """Write feed to file."""
        pass


class FeedFileWriterFactory:
    """Factory for creating specific feed writers."""

    def __init__(self) -> None:
        self._writers = {
            FeedType.JSON: JSONFeedFileWriter,
            FeedType.ATOM: AtomFeedFileWriter,
        }

    @property
    def feed_types(self) -> Set[FeedType]:
        """Set of feed types for which writers are registered."""
        return set(self._writers.keys())

    def create_writer(self, config: FeedFileWriterConfig) -> AbstractFeedFileWriter:
        """Create a feed writer for the specified feed type."""
        return self._writers[config.feed_type](config)


class JSONFeedFileWriter(AbstractFeedFileWriter):
    """Export feed as JSON-Feed format (https://www.jsonfeed.org/version/1.1/)."""

    async def write_feed(self, feed_meta: FeedMeta, feed_items: List[FeedItem]) -> None:
        """Write feed to JSON file."""

        feed: Dict[str, Any] = {
            "version": "https://jsonfeed.org/version/1.1",
            "title": feed_meta.title or "{} News".format(feed_meta.game.name.title()),
            "language": str(feed_meta.language),
            "home_page_url": "https://www.hoyolab.com/circles/{}".format(
                feed_meta.game
            ),
        }

        if self.config.url is not None:
            feed["feed_url"] = str(self.config.url)

        if feed_meta.icon is not None:
            feed["icon"] = str(feed_meta.icon)

        feed["items"] = [self.create_json_feed_item(item) for item in feed_items]

        try:
            async with aiofiles.open(self.config.path, "w", encoding="utf-8") as fd:
                await fd.write(json.dumps(feed))
        except IOError as err:
            raise FeedIOError(
                'Could not write JSON file to "{}"!'.format(self.config.path)
            ) from err

    @staticmethod
    def create_json_feed_item(item: FeedItem) -> Dict[str, Any]:
        """Convert FeedItem to JSON-Feed item."""

        json_item = {
            "id": str(item.id),
            "url": "https://www.hoyolab.com/article/{}".format(item.id),
            "title": item.title,
            "authors": [{"name": item.author}],
            "tags": [item.category.name.title()],
            "content_html": item.content,
            "date_published": item.published.astimezone().isoformat(),
        }

        if item.updated is not None:
            json_item["date_modified"] = item.updated.astimezone().isoformat()

        if item.image is not None:
            json_item["image"] = str(item.image)

        return json_item


class AtomFeedFileWriter(AbstractFeedFileWriter):
    """Export feed as Atom format (https://validator.w3.org/feed/docs/atom.html)."""

    async def write_feed(self, feed_meta: FeedMeta, feed_items: List[FeedItem]) -> None:
        """Write feed to Atom file."""

        # this is a workaround to avoid dealing with namespaces...
        meta_params = {
            "xmlns": "http://www.w3.org/2005/Atom",
            "xml:lang": str(feed_meta.language),
        }

        root = ElementTree.Element("feed", meta_params)

        id_str = "tag:hoyolab.com,2021:/official/{}".format(feed_meta.game)
        ElementTree.SubElement(root, "id").text = id_str

        title_str = feed_meta.title or "{} News".format(feed_meta.game.name.title())
        ElementTree.SubElement(root, "title").text = title_str

        updated_str = datetime.now().astimezone().isoformat()
        ElementTree.SubElement(root, "updated").text = updated_str

        ElementTree.SubElement(
            root,
            "link",
            {
                "href": "https://www.hoyolab.com/circles/{}".format(feed_meta.game),
                "rel": "alternate",
                "type": "text/html",
            },
        )

        if self.config.url:
            ElementTree.SubElement(
                root,
                "link",
                {
                    "href": self.config.url,
                    "rel": "self",
                    "type": "application/atom+xml",
                },
            )

        if feed_meta.icon:
            ElementTree.SubElement(root, "icon").text = feed_meta.icon

        entries = self.create_atom_feed_entries(feed_items)
        root.extend(entries)

        xml_bytes = ElementTree.tostring(root, encoding="utf-8", xml_declaration=True)

        try:
            async with aiofiles.open(self.config.path, "wb") as fd:
                await fd.write(xml_bytes)
        except IOError as err:
            raise FeedIOError(
                'Could not write Atom file to "{}"!'.format(self.config.path)
            ) from err

    @staticmethod
    def create_atom_feed_entries(
        feed_items: List[FeedItem],
    ) -> List[ElementTree.Element]:
        """Create Atom feed entries from given feed items."""

        entries = []

        for item in feed_items:
            entry = ElementTree.Element("entry")

            published_day = item.published.astimezone().date().isoformat()
            id_str = "tag:hoyolab.com,{}:{}".format(published_day, item.id)
            ElementTree.SubElement(entry, "id").text = id_str

            ElementTree.SubElement(entry, "title").text = item.title

            ElementTree.SubElement(
                entry,
                "link",
                {
                    "href": "https://www.hoyolab.com/article/{}".format(item.id),
                    "rel": "alternate",
                    "type": "text/html",
                },
            )

            ElementTree.SubElement(
                entry, "category", {"term": item.category.name.title()}
            )

            published_str = item.published.astimezone().isoformat()
            ElementTree.SubElement(entry, "published").text = published_str

            updated_str = (item.updated or item.published).astimezone().isoformat()
            ElementTree.SubElement(entry, "updated").text = updated_str

            author = ElementTree.SubElement(entry, "author")
            ElementTree.SubElement(author, "name").text = item.author

            ElementTree.SubElement(
                entry, "content", {"type": "html"}
            ).text = item.content

            entries.append(entry)

        return entries



from datetime import datetime
from enum import Enum
from enum import IntEnum, unique
from pathlib import Path
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar

from pydantic import BaseModel
from pydantic import HttpUrl

_IC = TypeVar("_IC", bound="FeedItemCategory")
_G = TypeVar("_G", bound="Game")


# --- ENUMS ---


@unique
class FeedItemCategory(IntEnum):
    NOTICES = 1
    EVENTS = 2
    INFO = 3

    @classmethod
    def from_str(cls: Type[_IC], category_str: str) -> _IC:
        try:
            return cls[category_str.upper()]
        except KeyError as err:
            raise ValueError('Unknown category "{}"!'.format(category_str)) from err


@unique
class Game(IntEnum):
    # 3 = unused, 5 = hoyolab, 7 = unused
    HONKAI = 1
    GENSHIN = 2
    THEMIS = 4
    STARRAIL = 6
    ZENLESS = 8

    @classmethod
    def from_str(cls: Type[_G], game_str: str) -> _G:
        try:
            return cls[game_str.upper()]
        except KeyError as err:
            raise ValueError('Unknown game "{}"!'.format(game_str)) from err


@unique
class FeedType(str, Enum):
    JSON = "json"
    ATOM = "atom"

    def __str__(self) -> str:  # pragma: no cover
        return self.value


@unique
class Language(str, Enum):
    GERMAN = "de-de"
    ENGLISH = "en-us"
    SPANISH = "es-es"
    FRENCH = "fr-fr"
    INDONESIAN = "id-id"
    ITALIAN = "it-it"
    JAPANESE = "ja-jp"
    KOREAN = "ko-kr"
    PORTUGUESE = "pt-pt"
    RUSSIAN = "ru-ru"
    THAI = "th-th"
    TURKISH = "tr-tr"
    VIETNAMESE = "vi-vn"
    CHINESE_CN = "zh-cn"
    CHINESE_TW = "zh-tw"

    def __str__(self) -> str:  # pragma: no cover
        return self.value


# --- PYDANTIC MODELS ---


class MyBaseModel(BaseModel):
    # https://docs.pydantic.dev/usage/model_config/#options
    class Config:
        extra = "forbid"
        anystr_strip_whitespace = True
        min_anystr_length = 1


class FeedMeta(MyBaseModel):
    game: Game
    category_size: int = 5
    categories: List[FeedItemCategory] = [c for c in FeedItemCategory]
    language: Language = Language.ENGLISH
    title: Optional[str] = None
    icon: Optional[HttpUrl] = None


class FeedItem(MyBaseModel):
    id: int
    title: str
    author: str
    content: str
    category: FeedItemCategory
    published: datetime
    updated: Optional[datetime] = None
    image: Optional[HttpUrl] = None


class FeedItemMeta(MyBaseModel):
    id: int
    last_modified: datetime


class FeedFileConfig(MyBaseModel):
    feed_type: FeedType
    path: Path


class FeedFileWriterConfig(FeedFileConfig):
    url: Optional[HttpUrl] = None


class FeedConfig(MyBaseModel):
    feed_meta: FeedMeta
    writer_configs: List[FeedFileWriterConfig]
    loader_config: Optional[FeedFileConfig] = None



import asyncio
import warnings
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar

import aiohttp

from .hoyolab import HoyolabNews
from .loaders import AbstractFeedFileLoader
from .loaders import FeedFileLoaderFactory
from .models import FeedConfig
from .models import FeedItem
from .models import FeedItemCategory
from .models import FeedMeta
from .writers import AbstractFeedFileWriter
from .writers import FeedFileWriterFactory

# used for class-methods
_GF = TypeVar("_GF", bound="GameFeed")
_GFC = TypeVar("_GFC", bound="GameFeedCollection")


class GameFeed:
    """Feed generator for a single game."""

    def __init__(
        self,
        feed_meta: FeedMeta,
        feed_writers: List[AbstractFeedFileWriter],
        feed_loader: Optional[AbstractFeedFileLoader] = None,
    ) -> None:
        # warn if identical paths for writers are found
        writer_paths = [str(writer.config.path) for writer in feed_writers]
        if len(writer_paths) != len(set(writer_paths)):
            warnings.warn(
                "Writers for {} game feed contain identical paths".format(
                    feed_meta.game.name.title()
                )
            )

        if feed_loader is None:
            loader_factory = FeedFileLoaderFactory()
            feed_loader = loader_factory.create_any_loader(feed_writers)

        self._feed_meta = feed_meta
        self._feed_writers = feed_writers
        self._feed_loader = feed_loader
        self._hoyolab = HoyolabNews(feed_meta.game, feed_meta.language)
        self._was_updated = False

    @property
    def was_updated(self) -> bool:
        """Flag if the feed has been updated after a create_feed() call."""
        return self._was_updated

    @classmethod
    def from_config(cls: Type[_GF], feed_config: FeedConfig) -> _GF:
        """Create an instance via a feed config."""

        writer_factory = FeedFileWriterFactory()
        writers = [
            writer_factory.create_writer(writer_config)
            for writer_config in feed_config.writer_configs
        ]

        loader_factory = FeedFileLoaderFactory()
        loader: AbstractFeedFileLoader
        if feed_config.loader_config:
            loader = loader_factory.create_loader(feed_config.loader_config)
        else:
            loader = loader_factory.create_any_loader(writers)

        return cls(feed_config.feed_meta, writers, loader)

    async def create_feed(
        self, session: Optional[aiohttp.ClientSession] = None
    ) -> None:
        """Create or update a feed and write it to files."""

        local_session = session or aiohttp.ClientSession()
        feed_categories = self._feed_meta.categories or [c for c in FeedItemCategory]
        self._was_updated = False

        feed_items = await self._feed_loader.get_feed_items()

        try:
            category_feeds = await asyncio.gather(
                *[
                    self._update_category_feed(
                        local_session,
                        category,
                        [item for item in feed_items if item.category == category],
                    )
                    for category in feed_categories
                ]
            )
        finally:
            if session is None:
                await local_session.close()

        if self._was_updated:
            combined_feed: List[FeedItem] = []
            for feed in category_feeds:
                combined_feed.extend(feed)

            # sort feed items descending by id -> latest at the top
            combined_feed.sort(key=lambda item: item.id, reverse=True)

            await asyncio.gather(
                *[
                    writer.write_feed(self._feed_meta, combined_feed)
                    for writer in self._feed_writers
                ]
            )

    async def _update_category_feed(
        self,
        session: aiohttp.ClientSession,
        category: FeedItemCategory,
        category_items: List[FeedItem],
    ) -> List[FeedItem]:
        """Create or update a specific category feed."""

        known_ids = {
            item.id: item.published
            if item.updated is None
            else max(item.published, item.updated)
            for item in category_items
            if item.category == category
        }

        latest_item_metas = await self._hoyolab.get_latest_item_metas(
            session, category, self._feed_meta.category_size
        )

        new_or_outdated_ids = {
            item_meta.id
            for item_meta in latest_item_metas
            if item_meta.id not in known_ids
            or item_meta.last_modified > known_ids[item_meta.id]
        }

        if len(new_or_outdated_ids) > 0:
            # remove outdated items from feed because they will be re-fetched
            category_items = list(
                filter(lambda item: item.id not in new_or_outdated_ids, category_items)
            )

            fetched_items = await asyncio.gather(
                *[
                    self._hoyolab.get_feed_item(session, item_id)
                    for item_id in new_or_outdated_ids
                ]
            )

            category_items.extend(fetched_items)

            # cut off older items that exceed category_size
            category_items.sort(key=lambda item: item.id, reverse=True)
            category_items = category_items[: self._feed_meta.category_size]

            self._was_updated = True

        return category_items


class GameFeedCollection:
    """Collection of feed generators for multiple games."""

    def __init__(
        self,
        feed_metas: List[FeedMeta],
        feed_writers: List[List[AbstractFeedFileWriter]],
        feed_loaders: List[Optional[AbstractFeedFileLoader]],
    ) -> None:
        if not (len(feed_metas) == len(feed_writers) == len(feed_loaders)):
            raise ValueError("Parameter lists do not have the same length!")

        self._game_feeds = [
            GameFeed(meta, writer, loader)
            for meta, writer, loader in zip(feed_metas, feed_writers, feed_loaders)
        ]

    @classmethod
    def from_configs(cls: Type[_GFC], feed_configs: List[FeedConfig]) -> _GFC:
        """Create an instance via feed configs."""

        metas: List[FeedMeta] = []
        writers: List[List[AbstractFeedFileWriter]] = []
        loaders: List[Optional[AbstractFeedFileLoader]] = []

        for feed_config in feed_configs:
            metas.append(feed_config.feed_meta)

            writer_factory = FeedFileWriterFactory()
            writers_configs = [
                writer_factory.create_writer(conf)
                for conf in feed_config.writer_configs
            ]
            writers.append(writers_configs)

            loader_factory = FeedFileLoaderFactory()
            loader = (
                loader_factory.create_loader(feed_config.loader_config)
                if feed_config.loader_config
                else None
            )
            loaders.append(loader)

        return cls(metas, writers, loaders)

    async def create_feeds(
        self, session: Optional[aiohttp.ClientSession] = None
    ) -> None:
        """Create or update a feed and write it to files."""

        local_session = session or aiohttp.ClientSession()

        try:
            await asyncio.gather(
                *[feed.create_feed(local_session) for feed in self._game_feeds]
            )
        finally:
            if session is None:
                await local_session.close()



import argparse
import asyncio
from pathlib import Path
from platform import system
from typing import Optional

from .configs import FeedConfigLoader
from .feeds import GameFeedCollection


async def create_feeds(config_path: Optional[Path] = None) -> None:
    # fallback path defined in config loader if no path given
    config_loader = FeedConfigLoader(config_path)

    if not config_loader.path.exists():
        await config_loader.create_default_config_file()
        print(
            'Default config file created at "{}"!'.format(config_loader.path.resolve())
        )
        return

    feed_configs = await config_loader.get_all_feed_configs()

    game_feed = GameFeedCollection.from_configs(feed_configs)
    await game_feed.create_feeds()


def cli() -> None:
    if system() == "Windows":
        # default policy not working on windows
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore

    arg_parser = argparse.ArgumentParser(
        prog="hoyolab-rss-feeds", description="Generate Hoyolab RSS feeds."
    )

    arg_parser.add_argument(
        "-c", "--config-path", help="path to the TOML config file", type=Path
    )

    args = arg_parser.parse_args()

    asyncio.run(create_feeds(args.config_path))


if __name__ == "__main__":
    cli()



from hoyolabrssfeeds import models
import pytest

# -- NOTE: This module does only test custom methods for pydantic models! --


def test_item_category_from_str(category: models.FeedItemCategory) -> None:
    cat = models.FeedItemCategory.from_str(category.name)
    assert cat == category


def test_invalid_category_str() -> None:
    with pytest.raises(ValueError):
        models.FeedItemCategory.from_str("Invalid")


def test_game_from_str(game: models.Game) -> None:
    g = models.Game.from_str(game.name)
    assert g == game


def test_invalid_game_str() -> None:
    with pytest.raises(ValueError):
        models.Game.from_str("Invalid")



import json
from platform import system
from stat import S_IREAD
from typing import List

import aiofiles
import atoma  # type: ignore
import pytest

from hoyolabrssfeeds import errors
from hoyolabrssfeeds import models
from hoyolabrssfeeds import writers


# ---- FACTORY TESTS ----


def test_factory_feed_types() -> None:
    factory = writers.FeedFileWriterFactory()
    expected = {models.FeedType.JSON, models.FeedType.ATOM}

    assert factory.feed_types == expected


def test_factory_create_writer(
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    factory = writers.FeedFileWriterFactory()
    json_writer = factory.create_writer(json_feed_file_writer_config)

    assert issubclass(type(json_writer), writers.AbstractFeedFileWriter)
    assert isinstance(json_writer, writers.JSONFeedFileWriter)

    atom_writer = factory.create_writer(atom_feed_file_writer_config)

    assert issubclass(type(atom_writer), writers.AbstractFeedFileWriter)
    assert isinstance(atom_writer, writers.AtomFeedFileWriter)


# ---- JSON WRITER TESTS ----


def test_json_feed_writer_config(
    json_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    writer = writers.JSONFeedFileWriter(json_feed_file_writer_config)

    assert writer.config == json_feed_file_writer_config


async def test_json_feed_writer(
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    feed_meta: models.FeedMeta,
    feed_item_list: List[models.FeedItem],
) -> None:
    writer = writers.JSONFeedFileWriter(json_feed_file_writer_config)

    await writer.write_feed(feed_meta, feed_item_list)

    assert json_feed_file_writer_config.path.exists()

    async with aiofiles.open(json_feed_file_writer_config.path, "r") as fd:
        feed_str = await fd.read()

    # this should raise an error if feed is invalid -> indirect feed validation
    feed = atoma.parse_json_feed(json.loads(feed_str))

    # just some example asserts
    assert len(feed.items) == len(feed_item_list)
    assert feed.title == feed_meta.title


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_write_json_feed_io_error(
    json_feed_file_writer_config: models.FeedFileWriterConfig,
    feed_meta: models.FeedMeta,
) -> None:
    writer = writers.JSONFeedFileWriter(json_feed_file_writer_config)

    # read-only file
    writer.config.path.touch(S_IREAD)

    with pytest.raises(errors.FeedIOError):
        await writer.write_feed(feed_meta, [])


# ---- ATOM WRITER TESTS ----


def test_atom_feed_writer_config(
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
) -> None:
    writer = writers.JSONFeedFileWriter(atom_feed_file_writer_config)

    assert writer.config == atom_feed_file_writer_config


async def test_atom_feed_writer(
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
    feed_meta: models.FeedMeta,
    feed_item_list: List[models.FeedItem],
) -> None:
    writer = writers.AtomFeedFileWriter(atom_feed_file_writer_config)

    await writer.write_feed(feed_meta, feed_item_list)

    assert atom_feed_file_writer_config.path.exists()

    async with aiofiles.open(atom_feed_file_writer_config.path, "rb") as fd:
        feed_bytes = await fd.read()

    # this should raise an error if feed is invalid -> indirect feed validation
    feed = atoma.parse_atom_bytes(feed_bytes)

    assert len(feed.entries) == len(feed_item_list)
    assert feed.title.value == feed_meta.title


@pytest.mark.skipif(system() == "Windows", reason="Currently not working on Windows")
async def test_write_atom_feed_io_error(
    atom_feed_file_writer_config: models.FeedFileWriterConfig,
    feed_meta: models.FeedMeta,
) -> None:
    writer = writers.AtomFeedFileWriter(atom_feed_file_writer_config)

    # read-only file
    writer.config.path.touch(S_IREAD)

    with pytest.raises(errors.FeedIOError):
        await writer.write_feed(feed_meta, [])
